@@ -3,7 +3,7 @@
  * Copyright (c) 1996 - 2000, Marek Michałkiewicz
  * Copyright (c) 2001       , Michał Moskal
  * Copyright (c) 2003 - 2005, Tomasz Kłoczko
- * Copyright (c) 2007 - 2009, Nicolas François
+ * Copyright (c) 2007 - 2013, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -33,7 +33,7 @@
 
 #include <config.h>
 
-#ident "$Id: pwmem.c 2777 2009-04-23 17:43:27Z nekral-guest $"
+#ident "$Id$"
 
 #include <stdio.h>
 #include "defines.h"
@@ -48,25 +48,53 @@
 	if (NULL == pw) {
 		return NULL;
 	}
-	*pw = *pwent;
+	/* The libc might define other fields. They won't be copied. */
+	memset (pw, 0, sizeof *pw);
+	pw->pw_uid = pwent->pw_uid;
+	pw->pw_gid = pwent->pw_gid;
+	/*@-mustfreeonly@*/
 	pw->pw_name = strdup (pwent->pw_name);
+	/*@=mustfreeonly@*/
 	if (NULL == pw->pw_name) {
+		free(pw);
 		return NULL;
 	}
+	/*@-mustfreeonly@*/
 	pw->pw_passwd = strdup (pwent->pw_passwd);
+	/*@=mustfreeonly@*/
 	if (NULL == pw->pw_passwd) {
+		free(pw->pw_name);
+		free(pw);
 		return NULL;
 	}
+	/*@-mustfreeonly@*/
 	pw->pw_gecos = strdup (pwent->pw_gecos);
+	/*@=mustfreeonly@*/
 	if (NULL == pw->pw_gecos) {
+		free(pw->pw_passwd);
+		free(pw->pw_name);
+		free(pw);
 		return NULL;
 	}
+	/*@-mustfreeonly@*/
 	pw->pw_dir = strdup (pwent->pw_dir);
+	/*@=mustfreeonly@*/
 	if (NULL == pw->pw_dir) {
+		free(pw->pw_gecos);
+		free(pw->pw_passwd);
+		free(pw->pw_name);
+		free(pw);
 		return NULL;
 	}
+	/*@-mustfreeonly@*/
 	pw->pw_shell = strdup (pwent->pw_shell);
+	/*@=mustfreeonly@*/
 	if (NULL == pw->pw_shell) {
+		free(pw->pw_dir);
+		free(pw->pw_gecos);
+		free(pw->pw_passwd);
+		free(pw->pw_name);
+		free(pw);
 		return NULL;
 	}
 
@@ -76,8 +104,10 @@
 void pw_free (/*@out@*/ /*@only@*/struct passwd *pwent)
 {
 	free (pwent->pw_name);
-	memzero (pwent->pw_passwd, strlen (pwent->pw_passwd));
-	free (pwent->pw_passwd);
+	if (pwent->pw_passwd) {
+		memzero (pwent->pw_passwd, strlen (pwent->pw_passwd));
+		free (pwent->pw_passwd);
+	}
 	free (pwent->pw_gecos);
 	free (pwent->pw_dir);
 	free (pwent->pw_shell);
